import sys

import django
import environ
from django.conf import settings

env = environ.Env()

db_url = env.db_url("DATABASE_URL")
# Django needs an explicit `postgis` engine, however,
# using `postgis` in the DSN is not supported by SQLAlchemy.
db_url["ENGINE"] = "django.contrib.gis.db.backends.postgis"


settings.configure(
    DATABASES={"default": db_url},
    DEBUG=True,
    INSTALLED_APPS=["schematools.contrib.django"],
    SCHEMA_URL=env.str("SCHEMA_URL"),
    AMSTERDAM_SCHEMA={"geosearch_disabled_datasets": ["bag"]},
    SCHEMA_DEFS_URL=env.str("SCHEMA_DEFS_URL", "https://schemas.data.amsterdam.nl/schema"),
)
django.setup()


def main() -> None:  # noqa: D103
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc
    execute_from_command_line(sys.argv)



# Generated by Django 3.1.8 on 2021-04-07 14:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0009_auto_20210330_1659"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dataset",
            name="schema_data",
            field=models.JSONField(verbose_name="Amsterdam Schema Contents"),
        ),
        migrations.AlterField(
            model_name="profile",
            name="schema_data",
            field=models.JSONField(verbose_name="Amsterdam Schema Contents"),
        ),
    ]



# Generated by Django 3.2.19 on 2023-07-09 13:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('datasets', '0015_alter_dataset_id_alter_datasetfield_id_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='dataset',
            name='view_data',
            field=models.TextField(blank=True, null=True, verbose_name='View SQL'),
        ),
    ]



from __future__ import annotations

import re
from functools import lru_cache
from typing import Final, Match, Pattern

from string_utils import slugify

from schematools import RELATION_INDICATOR

_RE_CAMEL_CASE: Final[Pattern[str]] = re.compile(
    r"(((?<=[^A-Z])[A-Z])|([A-Z](?![A-Z]))|((?<=[a-z])[0-9])|(?<=[0-9])[a-z]|((?<=[A-Z])[A-Z](?=[A-Z])))"  # noqa: E501
)

_CAMEL_CASE_REPLACE_PAT: Final[Pattern[str]] = re.compile(
    r"""
    (?:_|\s)+   # Find word boundaries by looking for underscore and whitespace characters, they
                # will be discarded (not captured)
    (.)         # Capture first letter of word on word boundary
    |           # OR
    (\d+)       # Capture a number
    (?:_|\s)*   # Optionally followed by underscore and whitespace characters (to be discarded)
    (.)         # Capture first letter of word on word boundary
    """,
    re.VERBOSE,
)


@lru_cache(maxsize=500)
def toCamelCase(ident: str, first_upper=False) -> str:
    """Convert an identifier to camelCase format.

    Word boundaries are determined by:
    - numbers
    - underscore characters
    - whitespace characters (this violates the concept of identifiers,
      but we handle it nevertheless)

    A camelCased identifier, when it starts with a letter, it will start with a lower cased letter.

    Empty strings are not allowed. They will raise an :exc:`ValueError` exception.

    Examples::

        >>> toCamelCase("dataset_table_schema")
        'datasetTableSchema'
        >>> toCamelCase("dataset table schema")
        'datasetTableSchema'
        >>> toCamelCase("fu_33_bar")
        'fu33Bar'
        >>> toCamelCase("fu_33bar")
        'fu33Bar'
        >>> toCamelCase("fu_33Bar")
        'fu33Bar'
        >>> toCamelCase("33_fu_bar")
        '33FuBar'

    Args:
        ident: The identifier to be converted.

    Returns:
        The identifier in camelCase format.

    Raises:
        ValueError: If ``indent`` is an empty string.

    """
    if ident == "":
        raise ValueError("Parameter `ident` cannot be an empty string.")

    def replacement(m: Match) -> str:
        # As we use the OR operator in the regular expression with capture groups on both sides,
        # we will always have at least one capture group that results in `None`. We filter those
        # out in the generator expression. Even though a captured group sometimes represents a
        # number (as a string), we still call `upper()` on it. That's faster than another
        # explicit test.
        return "".join(s.upper() for s in m.groups() if s)

    result = _CAMEL_CASE_REPLACE_PAT.sub(replacement, ident)
    if first_upper:
        return result[0].upper() + result[1:]
    else:
        # The first letter of camelCase identifier is always lower case
        return result[0].lower() + result[1:]


@lru_cache(maxsize=500)
def to_snake_case(ident: str) -> str:
    """Convert an identifier to snake_case format.

    Empty strings are not allowed. They will raise an :exc:`ValueError` exception.

    Args:
        ident: The identifier to be converted.

    Returns:
        The identifier in snake_case foramt.

    Raises:
        ValueError: If ``ident`` is an empty string.
    """
    if ident == "":
        raise ValueError("Parameter `ident` cannot be an empty string.")
    # Convert to field name, avoiding snake_case to snake_case issues.
    # Also preserve RELATION_INDICATOR in names (RELATION_INDICATOR are used for object relations)
    name_parts = [toCamelCase(part) for part in ident.split(RELATION_INDICATOR)]
    return RELATION_INDICATOR.join(
        slugify(_RE_CAMEL_CASE.sub(r" \1", part).strip(), separator="_") for part in name_parts
    )



import operator

import pytest

from schematools.exceptions import SchemaObjectNotFound
from schematools.types import (
    DatasetSchema,
    DatasetTableSchema,
    Permission,
    PermissionLevel,
    ProfileSchema,
    SemVer,
)


def test_permission_level_ordering() -> None:
    """Test whether enum ordering works based on the int values."""
    assert sorted(PermissionLevel) == [
        PermissionLevel.NONE,
        PermissionLevel.SUBOBJECTS_ONLY,
        PermissionLevel.LETTERS,
        PermissionLevel.RANDOM,
        PermissionLevel.ENCODED,
        PermissionLevel.READ,
    ]
    assert PermissionLevel.highest is PermissionLevel.READ
    assert PermissionLevel.highest is max(PermissionLevel)


def test_geo_and_id_when_configured(schema_loader, meetbouten_schema) -> None:
    table = meetbouten_schema.get_table_by_id("meetbouten")
    assert table.identifier == ["identificatie"]
    assert table.main_geometry == "geometrie"
    id_field = [field for field in table.fields if [field.name] == table.identifier][0]
    assert id_field.is_primary


def test_geo_and_id_when_not_configured(schema_loader, afvalwegingen_schema) -> None:
    table = afvalwegingen_schema.get_table_by_id("containers")
    assert table.identifier == ["id"]
    assert table.main_geometry == "geometry"
    id_field = [field for field in table.fields if [field.name] == table.identifier][0]
    assert id_field.is_primary


def test_import_dataset_separate_table_files(schema_loader) -> None:
    """Prove that datasets with tables in separate files are created correctly."""
    schema = schema_loader.get_dataset("gebieden_sep_tables")
    assert len(schema.tables) == 2
    table = schema.get_table_by_id("buurten")
    assert table.main_geometry == "primaireGeometrie"


def test_datasetschema_from_file_not_a_dataset(schema_loader) -> None:
    """Ensure a proper exception is raised when loading a file that's not a DatasetSchema."""

    error_msg = "Invalid Amsterdam Dataset schema file"
    with pytest.raises(ValueError, match=error_msg):
        # v1.0.0.json is a DatasetRow, not a DatasetSchema.
        schema_loader.get_dataset_from_file("gebieden_sep_tables/bouwblokken/v1.0.0.json")

    error_msg = "Invalid JSON file"
    with pytest.raises(ValueError, match=error_msg):
        # not_a_json_file.txt is not a JSON file. We should still get our ValueError.
        schema_loader.get_dataset_from_file("not_a_json_file.txt")


def test_profile_schema(brp_r_profile_schema: ProfileSchema) -> None:
    """Prove that the profile files are properly read,
    and have their fields access the JSON data.
    """
    assert brp_r_profile_schema.scopes == {"BRP/R"}

    brp = brp_r_profile_schema.datasets["brp"]
    table = brp.tables["ingeschrevenpersonen"]

    assert table.permissions.level is PermissionLevel.READ
    assert table.fields["bsn"] == Permission(PermissionLevel.ENCODED)
    assert table.mandatory_filtersets == [
        ["bsn", "lastname"],
        ["postcode", "lastname"],
    ]


def test_fetching_of_related_schema_ids(schema_loader) -> None:
    """Prove that ids of related dataset schemas are properly collected."""
    schema = schema_loader.get_dataset_from_file("multirelation.json")
    assert set(schema.related_dataset_schema_ids) == {"gebieden", "baseDataset"}


def test_semver_init() -> None:
    """Test SemVer initialization (including raising ValueError's)"""
    sv = SemVer("1.2.3")
    assert sv.major == 1
    assert sv.minor == 2
    assert sv.patch == 3

    sv_no_patch = SemVer("1.2")
    assert sv_no_patch.major == 1
    assert sv_no_patch.minor == 2
    assert sv_no_patch.patch == 0  # default when not explicitly specified

    sv_no_minor = SemVer("1")
    assert sv_no_minor.major == 1
    assert sv_no_minor.minor == 0  # default when not explicitly specified
    assert sv_no_minor.patch == 0  # default when not explicitly specified

    sv_with_v_prefix = SemVer("v1.2.3")
    assert sv_with_v_prefix.major == 1
    assert sv_with_v_prefix.minor == 2
    assert sv_with_v_prefix.patch == 3

    invalid_semver_values = ("1.0.0.0", "-1.0.0", "fubar")
    for isv in invalid_semver_values:
        with pytest.raises(ValueError):
            SemVer(isv)


def test_semver_str() -> None:
    """Test SemVer str representation."""
    assert str(SemVer("1.2.3")) == "1.2.3"
    assert str(SemVer("1.2")) == "1.2.0"
    assert str(SemVer("1")) == "1.0.0"

    assert str(SemVer("v1.2.3")) == "1.2.3"
    assert str(SemVer("v1.2")) == "1.2.0"
    assert str(SemVer("v1")) == "1.0.0"


def test_semver_repr() -> None:
    """Test SemVer repr representation."""
    assert repr(SemVer("1.2.3")) == 'SemVer("1.2.3")'
    assert repr(SemVer("1.2")) == 'SemVer("1.2.0")'
    assert repr(SemVer("1")) == 'SemVer("1.0.0")'

    assert repr(SemVer("v1.2.3")) == 'SemVer("1.2.3")'
    assert repr(SemVer("v1.2")) == 'SemVer("1.2.0")'
    assert repr(SemVer("v1")) == 'SemVer("1.0.0")'


def test_semver_lt() -> None:
    assert not SemVer("v1.0.0") < SemVer("1.0.0")
    assert not SemVer("1.0.0") < SemVer("v1.0.0")
    assert SemVer("v1.0.0") < SemVer("1.0.2")
    assert SemVer("v1.0.0") < SemVer("1.2.0")


def test_semver_le() -> None:
    assert SemVer("v1.0.0") <= SemVer("1.0.0")
    assert SemVer("1.0.0") <= SemVer("v1.0.0")
    assert SemVer("v1.0.0") <= SemVer("1.0.2")
    assert SemVer("v1.0.0") <= SemVer("1.2.0")


def test_semver_gt() -> None:
    assert not SemVer("v1.0.0") > SemVer("1.0.0")
    assert not SemVer("1.0.0") > SemVer("v1.0.0")
    assert SemVer("v1.0.2") > SemVer("1.0.0")
    assert SemVer("v1.2.0") > SemVer("1.0.0")


def test_semver_ge() -> None:
    assert SemVer("v1.0.0") >= SemVer("1.0.0")
    assert SemVer("1.0.0") >= SemVer("v1.0.0")
    assert SemVer("v1.0.2") >= SemVer("1.0.0")
    assert SemVer("v1.2.0") >= SemVer("1.0.0")


def test_semver_eq() -> None:
    assert SemVer("v1.0.0") == SemVer("1.0.0")
    assert SemVer("1.0.0") == SemVer("v1.0.0")
    assert SemVer("1.0.2") != SemVer("1.0.0")
    assert SemVer("1.2.0") != SemVer("1.0.0")


def test_dataset_schema_get_fields_with_surrogate_pk(
    composite_key_schema: DatasetSchema, verblijfsobjecten_schema: DatasetSchema
):
    """Prove that the surrogate 'id' key is returned once for schemas with a
    composite key, regardless of whether the surrogate key is already defined
    by the schema or generated"""

    verblijfsobjecten = verblijfsobjecten_schema.tables[0]
    composite_key_schema = composite_key_schema.tables[0]

    # this schema gets a generated 'id'
    assert sorted(x.id for x in verblijfsobjecten.get_fields()) == [
        "beginGeldigheid",
        "eindGeldigheid",
        "gebruiksdoel",
        "id",
        "identificatie",
        "ligtInBuurt",
        "schema",
        "volgnummer",
    ]

    # this schema defines an 'id'
    assert sorted(x.name for x in composite_key_schema.get_fields()) == [
        "beginGeldigheid",
        "eindGeldigheid",
        "id",
        "identificatie",
        "schema",
        "volgnummer",
    ]


def test_dataset_with_loose_1n_relations_has_no_through_tables(meldingen_schema, gebieden_schema):
    """Prove that a loose relation for a 1-N is not generating a though table.

    The `meldingen_schema` has a 1-N relation to `buurt`, however,
    that is a loose relation. Those relations should not get a through tables,
    so, the only one table that should be generated is the main meldingen table.
    """
    tables_including_through = meldingen_schema.get_tables(include_through=True)
    assert len(tables_including_through) == 1


def test_dataset_with_loose_nm_relations_has_through_tables(
    woningbouwplannen_schema, gebieden_schema
):
    """Prove that a loose relation for a NM relation is generating a though table.

    The `woningbouwplannen_schema` has two NM relations to `buurt`, however,
    those are loose relations. Those relations should get a through tables,
    because NM relations always need a through table by nature.
    So, in total, 3 tables should be created.
    """
    tables_including_through = woningbouwplannen_schema.get_tables(include_through=True)
    assert len(tables_including_through) == 3


def test_dataset_with_camel_cased_id_generates_correct_through_relations(
    schema_loader, gebieden_schema
):
    """Prove that the relation identifiers in through tables are correctly generated.

    When a through table is generated, two FK relations are inserted in this table,
    one to the source and one to the target table.
    When dataset.id or table.id in the source or target table are camelCased,
    the generated identifiers should stay intact, and should not be snakecased.
    """
    schema = schema_loader.get_dataset_from_file("multirelation.json")
    through_table = schema.through_tables[0]

    # The fields `hasrelations` and `hasNMRelation` are the FK's to source and target table.
    assert through_table.get_field_by_id("hasrelations").relation == "baseDataset:hasrelations"
    assert through_table.get_field_by_id("hasNMRelation").relation == "baseDataset:internalRelated"


def test_subfields(ggwgebieden_schema: DatasetSchema) -> None:
    field = ggwgebieden_schema.get_table_by_id("buurten").get_field_by_id("ligtInWijk")
    subfields = sorted(field.subfields, key=operator.attrgetter("id"))
    assert subfields[0].id == "identificatie"
    assert subfields[1].id == "volgnummer"

    assert subfields[0] == field.get_field_by_id("identificatie")
    assert subfields[1] == field.get_field_by_id("volgnummer")
    with pytest.raises(SchemaObjectNotFound):
        field.get_field_by_id("iDoNotExist")


def test_names_of_subobject_fields(kadastraleobjecten_schema: DatasetSchema) -> None:
    """Prove that the subfields of an object field get prefixed."""
    field = kadastraleobjecten_schema.get_table_by_id("kadastraleobjecten").get_field_by_id(
        "soortCultuurOnbebouwd"
    )
    assert {"soortCultuurOnbebouwdCode", "soortCultuurOnbebouwdOmschrijving"} == {
        sf.name for sf in field.subfields
    }


def test_json_subfield_does_not_crash(kadastraleobjecten_schema: DatasetSchema) -> None:
    """Prove that the subfields of an object field get prefixed."""
    field = kadastraleobjecten_schema.get_table_by_id("kadastraleobjecten").get_field_by_id(
        "soortGrootte"
    )
    assert len(field.subfields) == 0


def test_raise_exception_on_missing_properties_in_array(schema_loader):
    """Test if a human-readable error is raised on missing properties key."""
    schema = schema_loader.get_dataset_from_file("missing_properties.json")
    with pytest.raises(
        KeyError, match=r"Key 'properties' not defined in 'meetbouten.broken_array'"
    ):
        schema.get_table_by_id("meetbouten")


def test_load_publisher_object_from_dataset(schema_loader):
    """Test that we can retrieve a publisher object from a DatasetSchema
    as defined by metaschema 2.0"""
    schema = schema_loader.get_dataset_from_file("metaschema2.json")

    assert schema.publisher == {
        "name": "Datateam Harry",
        "id": "HARRY",
        "shortname": "harhar",
        "tags": {"team": "taggy", "costcenter": "123456789.4321.13519"},
    }


def test_repr_broken_schema():
    """Regression test: __repr__ and __missing__ performed infinite mutual recursion
    when dealing with broken schemas.
    """
    try:
        DatasetTableSchema({}, parent_schema=None)
    except KeyError:  # KeyError is ok, RecursionError isn't.
        pass


def test_relation_with_extra_properties_has_through_table(gebieden_schema):
    """Prove that a 1-N relation with extra properties on the relates add a through table."""
    tables_including_through = {t.id for t in gebieden_schema.get_tables(include_through=True)}
    assert "ggwgebieden_ligtInStadsdeel" in tables_including_through


def test_extra_properties_of_relation_field_are_also_in_through_table(gebieden_schema):
    """Prove that extra fields defined on the relation are also showing up in the through table."""
    tables_including_through = {t.id: t for t in gebieden_schema.get_tables(include_through=True)}
    fields_on_through_table = {
        f.id: f for f in tables_including_through["bouwblokken_ligtInBuurt"].fields
    }
    # the field `ligtInBuurt` that is on the through table should have the fields
    # `beginGeldigheid` and `eindGeldigheid` because those are also defined on the field
    # in the source table of the relation.
    assert {"beginGeldigheid", "eindGeldigheid"} < fields_on_through_table[
        "ligtInBuurt"
    ].json_data()["properties"].keys()



from schematools.importer.base import Provenance


def test_row_plain():
    """Prove that a regular dict just work as is."""
    value = Provenance("colname1").resolve({"colname1": 12, "colname2": "test"})
    assert value == 12


def test_row_with_jsonpath_provenance():
    """Prove that a provenance based on json path works."""
    value = Provenance("$.colname2.sub").resolve({"colname1": 12, "colname2": {"sub": "test"}})
    assert value == "test"



import enum
import typing
from dataclasses import InitVar, dataclass, field


class LayerType(str, enum.Enum):
    point = "POINT"
    polygon = "POLYGON"
    # TODO add otehr layer types here


class Metadata(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self["__type__"] = "metadata"


@dataclass
class Connection:
    type: str
    data: str

    def __init__(self, type, s):
        self.type = type
        self.data = s

    @classmethod
    def for_postgres(cls, user, pw, dbname, host):
        return cls("postgis", f"user={user} password={pw} dbname={dbname} host={host}")

    def __str__(self):
        return self.data


class Data(str):
    @classmethod
    def for_postgres(cls, column, table, srid=None, UNIQUE=None):
        result = cls(f"{column} from {table}")
        if srid:
            result += f" USING srid={srid}"
        if UNIQUE:
            result += f" USING UNIQUE {UNIQUE}"
        return result


@dataclass
class FeatureClass:
    """Used for rendering a feature"""

    name: typing.Optional[str] = None
    expression: typing.Optional[str] = None
    # https://mapserver.org/mapfile/style.html#style
    styles: typing.List[dict] = field(default_factory=list)
    # https://mapserver.org/mapfile/label.html#label
    labels: typing.List[dict] = field(default_factory=list)

    __type__: str = field(init=False, default="class")

    def add_style(self, d):
        self.styles.append({"__type__": "style", **d})

    def add_label(self, d):
        self.labels.append({"__type__": "label", **d})


Filename = str


@dataclass
class Layer:
    name: str
    type: str  # TODO: get this type from schema
    with_connection: InitVar[Connection] = None
    projection: typing.Optional[typing.List[str]] = None

    connection: typing.Optional[str] = field(init=False, default=None)
    connectiontype: typing.Optional[str] = field(init=False, default=None)

    data: typing.List[Data] = field(default_factory=list)
    classes: typing.List[FeatureClass] = field(default_factory=list)

    include: typing.List[Filename] = field(default_factory=list)
    labelitem: typing.Optional[str] = None
    metadata: Metadata = field(default_factory=Metadata)

    __type__: str = field(init=False, default="layer")

    def __post_init__(self, connection: Connection = None):
        if connection:
            self.connection = connection.data
            self.connectiontype = connection.type


@dataclass
class Web:
    metadata: Metadata = field(default_factory=Metadata)
    __type__: str = field(init=False, default="web")


@dataclass
class Mapfile:
    name: str
    layers: typing.List[Layer]
    projection: typing.Optional[typing.List[str]] = None
    include: typing.List[Filename] = field(default_factory=list)
    web: typing.Optional[Web] = None



import pytest

from schematools.naming import to_snake_case, toCamelCase


def test_toCamelCase() -> None:
    """Confirm that:
    - space separated name is converted to camelCase
    - PascalCase results in camelCase
    - snake_case results in camelCase
    """
    assert toCamelCase("test name magic") == "testNameMagic"
    assert toCamelCase("test name magic2") == "testNameMagic2"
    assert toCamelCase("testNameMagic") == "testNameMagic"
    assert toCamelCase("TestNameMagic") == "testNameMagic"
    assert toCamelCase("test_name_magic") == "testNameMagic"
    assert toCamelCase("numbers_33_in_the_middle_44") == "numbers33InTheMiddle44"
    # mind the lower case "i" after "33". It should be made upper case
    assert toCamelCase("numbers33inTheMiddle44") == "numbers33InTheMiddle44"
    assert toCamelCase("per_jaar_per_m2") == "perJaarPerM2"
    assert toCamelCase("eerste_h_n_id") == "eersteHNId"

    with pytest.raises(ValueError):
        toCamelCase("")


def test_to_snake_case() -> None:
    """Confirm that:
    - space separated name converted to snake_case
    - camelCase converted to snake_case
    - snake_case converted to snake_case
    """
    assert to_snake_case("test name magic") == "test_name_magic"
    assert to_snake_case("test name magic22") == "test_name_magic_22"
    assert to_snake_case("TestNameMagic") == "test_name_magic"
    assert to_snake_case("testNameMagic") == "test_name_magic"
    assert to_snake_case("test_name_magic") == "test_name_magic"
    assert to_snake_case("hoofdroutes_u_routes") == "hoofdroutes_u_routes"
    assert to_snake_case("verlengingSluitingstijd1") == "verlenging_sluitingstijd_1"
    assert to_snake_case("numbers33inTheMiddle44") == "numbers_33_in_the_middle_44"
    assert to_snake_case("perJaarPerM2") == "per_jaar_per_m2"
    assert to_snake_case("eersteHNId") == "eerste_h_n_id"

    with pytest.raises(ValueError):
        to_snake_case("")


def test_snake_and_camel_case_functions_are_inverses() -> None:
    """Prove that snake and camel case functions are inverses of eachother."""
    # mind the lower case "i" after "33". It should be made upper case
    assert to_snake_case(toCamelCase("per_jaar_per_m2")) == "per_jaar_per_m2"
    assert to_snake_case(toCamelCase("test_name_magic")) == "test_name_magic"
    assert (
        to_snake_case(toCamelCase("numbers_33_in_the_middle_44")) == "numbers_33_in_the_middle_44"
    )
    assert to_snake_case(toCamelCase("eerste_h_n_id")) == "eerste_h_n_id"
    assert toCamelCase(to_snake_case("numbers33InTheMiddle44")) == "numbers33InTheMiddle44"

    assert toCamelCase(to_snake_case("testNameMagic")) == "testNameMagic"
    assert toCamelCase(to_snake_case("testNameMagic2")) == "testNameMagic2"
    assert toCamelCase(to_snake_case("eersteHNId")) == "eersteHNId"



from django.apps import AppConfig
from django.db import models

# Make sure OneToOneField also allows __contains= lookups (will only work for type=string).
# This allows model_factory() to create the CheckConstraint check for a relational PK field.
models.OneToOneField.register_lookup(models.lookups.Contains)


class SchematoolsAppConfig(AppConfig):
    name = "schematools.contrib.django"
    # Alias as `datasets` app to avoid unnecessary migrations on legacy systems.
    label = "datasets"
    default_auto_field = "django.db.models.BigAutoField"



# Generated by Django 4.1.7 on 2024-01-10 16:59

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("datasets", "0016_dataset_view_data"),
    ]

    operations = [
        migrations.AddField(
            model_name="dataset",
            name="enable_export",
            field=models.BooleanField(default=False),
        ),
    ]



from typing import Any, Iterable

from django.core.management import BaseCommand, CommandError, CommandParser

from schematools.contrib.django.models import Dataset


class BaseDatasetCommand(BaseCommand):
    """A management command that works on the list of imported datasets."""

    def add_arguments(self, parser: CommandParser) -> None:
        """Provide default arguments to pass dataset names to this management command."""
        parser.add_argument("dataset", nargs="*", help="Names of the datasets.")
        parser.add_argument(
            "-x",
            "--exclude",
            dest="exclude",
            nargs="*",
            default=[],
            help="Datasets that need to be skipped.",
        )

    def get_datasets(
        self, options: dict[str, Any], enable_db=None, default_all=False
    ) -> Iterable[Dataset]:
        """Provide the datasets based on the command options"""
        datasets = Dataset.objects.all()
        if enable_db is not None:
            datasets = datasets.filter(enable_db=enable_db)

        if options["dataset"]:
            names = set(options["dataset"]) - set(options["exclude"])
            datasets = datasets.filter(name__in=names)
            if invalid_names := names - {ds.name for ds in datasets}:
                raise CommandError(f"Datasets not found: {', '.join(sorted(invalid_names))}")
        elif options["exclude"]:
            datasets = datasets.exclude(name__in=options["exclude"])
        elif not default_all:
            raise CommandError("Provide at least a dataset by name, or use the --exclude option.")

        return datasets



from __future__ import annotations

import logging
from typing import List

from django.db import connection
from django.db.models import sql

from schematools.contrib.django.factories import schema_model_mockers_factory
from schematools.contrib.django.models import Dataset, DynamicModel

logger = logging.getLogger(__name__)


def create_data_for(
    *datasets: Dataset,
    start_at: int = 1,
    size: int = 50,
    sql: bool = False,
    tables: List[str] | None = None,
) -> List[str]:
    """Create mock data for the indicated datasets."""
    limit_tables_to = set(tables) if tables is not None else set()
    for dataset in datasets:
        if not dataset.enable_db:
            logger.warning("Skipping `%s`, `enable_db` is False", dataset.name)
            continue
        model_mockers = {
            cls._meta.get_model_class()._meta.model_name: cls
            for cls in schema_model_mockers_factory(dataset, base_app_name="dso_api.dynamic_api")
        }

        for mock_model in model_mockers.values():
            if limit_tables_to and mock_model._meta.model.__name__ not in limit_tables_to:
                continue
            if start_at > 1:
                mock_model._setup_next_sequence = lambda: start_at
            if sql:
                return list(_get_sql_for(mock_model.build_batch(size)))
            else:
                mock_model.create_batch(size)


def _get_sql_for(objects: List[DynamicModel]):
    """Get the SQL insert statements for the provided model objects."""
    # We need a real cursor here, so that `cursor.mogrify`
    # knows exactly how to render the query.
    cursor = connection.cursor()
    for obj in objects:
        values = obj._meta.local_fields
        query = sql.InsertQuery(obj)
        query.insert_values(values, [obj])
        compiler = query.get_compiler("default")
        statements = compiler.as_sql()
        for statement, params in statements:
            yield cursor.mogrify(statement, params).decode() + ";"



# Generated by Django 3.0.3 on 2020-02-26 07:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0003_datasettable_geometry_field_type"),
    ]

    operations = [
        migrations.AddField(
            model_name="dataset",
            name="auth",
            field=models.CharField(
                blank=True, max_length=250, null=True, verbose_name="Authorization"
            ),
        ),
        migrations.AddField(
            model_name="datasettable",
            name="auth",
            field=models.CharField(blank=True, max_length=250, null=True),
        ),
    ]



"""Event tests."""
import json
from datetime import date, datetime

import orjson
import pytest
from more_itertools import peekable

from schematools.events.full import EventsProcessor

# pytestmark = pytest.mark.skip("all tests disabled")


def _load_events_from_file_as_full_load_sequence(importer, events_path):
    with open(events_path, "rb") as ef:
        is_first = True
        p = peekable(ef)

        for line in p:
            event_id, event_meta_str, data_str = line.split(b"|", maxsplit=2)
            event_meta = orjson.loads(event_meta_str)
            event_meta |= {
                "full_load_sequence": True,
                "first_of_sequence": is_first,
                "last_of_sequence": p.peek(None) is None,
            }
            event_data = orjson.loads(data_str)
            importer.process_event(
                event_meta,
                event_data,
            )
            is_first = False


def test_event_process_insert(
    here, db_schema, tconn, local_metadata, gebieden_schema, benk_schema
):
    """Prove that event gets inserted."""
    events_path = here / "files" / "data" / "bouwblokken.gobevents"
    importer = EventsProcessor(
        [gebieden_schema], tconn, local_metadata=local_metadata, truncate=True
    )
    importer.load_events_from_file(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM gebieden_bouwblokken")]
    assert len(records) == 2
    assert records[0]["code"] == "AA01"
    assert records[1]["code"] == "AA02"
    assert records[0]["eind_geldigheid"] is None
    assert records[0]["begin_geldigheid"] == date(2006, 6, 12)


def test_event_process_insert_object(
    here, db_schema, tconn, local_metadata, nap_schema, gebieden_schema
):
    """Prove that event gets inserted correctly with object split in columns."""
    events_path = here / "files" / "data" / "peilmerken.gobevents"
    importer = EventsProcessor(
        [nap_schema, gebieden_schema], tconn, local_metadata=local_metadata, truncate=True
    )
    importer.load_events_from_file(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]
    assert records[0]["status_code"] == 3
    assert records[0]["status_omschrijving"] == "Vervallen"


def test_event_process_update(here, tconn, local_metadata, gebieden_schema):
    """Prove that event gets updated."""
    events_path = here / "files" / "data" / "bouwblokken_update.gobevents"
    importer = EventsProcessor(
        [gebieden_schema], tconn, local_metadata=local_metadata, truncate=True
    )
    importer.load_events_from_file(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM gebieden_bouwblokken")]
    assert len(records) == 1
    assert records[0]["code"] == "AA01"
    assert records[0]["begin_geldigheid"] == date(2020, 2, 5)
    assert records[0]["registratiedatum"] == datetime(2020, 2, 5, 15, 6, 43)


def test_event_process_delete(here, tconn, local_metadata, gebieden_schema):
    """Prove that event gets deleted."""
    events_path = here / "files" / "data" / "bouwblokken_delete.gobevents"
    importer = EventsProcessor(
        [gebieden_schema], tconn, local_metadata=local_metadata, truncate=True
    )
    importer.load_events_from_file(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM gebieden_bouwblokken")]
    assert len(records) == 1
    assert records[0]["code"] == "AA01"


def test_event_process_nm_relation_delete(here, tconn, local_metadata, gebieden_schema, salogger):
    """Prove that NM relations of an event get deleted."""
    events_path = (
        here / "files" / "data" / "gebieden_ggwgebieden_bestaat_uit_buurten_delete.gobevents"
    )
    importer = EventsProcessor(
        [gebieden_schema], tconn, local_metadata=local_metadata, truncate=True
    )
    importer.load_events_from_file(events_path)
    records = [
        dict(r) for r in tconn.execute("SELECT * FROM gebieden_ggwgebieden_bestaat_uit_buurten")
    ]
    assert len(records) == 0


def test_event_process_relation_update_parent_table(
    here, db_schema, tconn, local_metadata, nap_schema, gebieden_schema
):
    events_path = here / "files" / "data" / "peilmerken.gobevents"
    importer = EventsProcessor([nap_schema, gebieden_schema], tconn, local_metadata=local_metadata)
    importer.load_events_from_file_using_bulk(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]

    # Imported objects without relations
    assert len(records) == 1
    assert records[0]["ligt_in_bouwblok_id"] is None
    assert records[0]["ligt_in_bouwblok_identificatie"] is None
    assert records[0]["ligt_in_bouwblok_volgnummer"] is None

    events_path = here / "files" / "data" / "peilmerken_ligt_in_bouwblok.gobevents"
    importer = EventsProcessor([nap_schema, gebieden_schema], tconn, local_metadata=local_metadata)
    importer.load_events_from_file_using_bulk(events_path)
    rel_records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken_ligt_in_bouwblok")]
    parent_records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]

    # Should have inserted the rel and updated relation columns in  parent (object) table
    assert len(rel_records) == 1
    assert len(parent_records) == 1

    assert parent_records[0]["ligt_in_bouwblok_id"] == "03630012095746.1"
    assert parent_records[0]["ligt_in_bouwblok_identificatie"] == "03630012095746"
    assert parent_records[0]["ligt_in_bouwblok_volgnummer"] == 1

    events_path = here / "files" / "data" / "peilmerken_ligt_in_bouwblok.delete.gobevents"
    importer = EventsProcessor([nap_schema, gebieden_schema], tconn, local_metadata=local_metadata)
    importer.load_events_from_file_using_bulk(events_path)
    rel_records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken_ligt_in_bouwblok")]
    parent_records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]

    # Rel table row should be deleted, fields in parent table should be set to None again
    assert len(rel_records) == 0
    assert len(parent_records) == 1
    assert records[0]["ligt_in_bouwblok_id"] is None
    assert records[0]["ligt_in_bouwblok_identificatie"] is None
    assert records[0]["ligt_in_bouwblok_volgnummer"] is None


def test_event_process_relation_skip_update_parent_table_nm_relations(
    here, db_schema, tconn, local_metadata, gebieden_schema
):
    events_path = here / "files" / "data" / "gebieden_ggwgebieden_bestaat_uit_buurten.gobevents"
    importer = EventsProcessor([gebieden_schema], tconn, local_metadata=local_metadata)

    # First import row in parent table
    importer.load_events_from_file(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM gebieden_ggwgebieden")]
    assert len(records) == 1

    # Then import row in relation table and check that parent table is not updated and no error is
    # raised
    events_path = (
        here / "files" / "data" / "gebieden_ggwgebieden_bestaat_uit_buurten_reltable_add.gobevents"
    )
    importer.load_events_from_file(events_path)
    records = [
        dict(r) for r in tconn.execute("SELECT * FROM gebieden_ggwgebieden_bestaat_uit_buurten")
    ]
    assert len(records) == 1
    assert records[0]["ggwgebieden_id"] == "03630950000000.1"
    assert records[0]["bestaat_uit_buurten_id"] == "03630023754008ADD.1"

    # Update row in relation table and check that parent table is not updated and no error is
    # raised
    events_path = (
        here
        / "files"
        / "data"
        / "gebieden_ggwgebieden_bestaat_uit_buurten_reltable_modify.gobevents"
    )
    importer.load_events_from_file(events_path)
    records = [
        dict(r) for r in tconn.execute("SELECT * FROM gebieden_ggwgebieden_bestaat_uit_buurten")
    ]
    assert len(records) == 1
    assert records[0]["ggwgebieden_id"] == "03630950000000.1"
    assert records[0]["bestaat_uit_buurten_id"] == "03630023754008MOD.1"


def test_event_process_relation_update_parent_table_shortname(
    here, db_schema, tconn, local_metadata, gebieden_schema
):
    """Tests updating of the parent table with a relation attribute with a shortname."""

    # Import bouwblokken in object table
    events_path = here / "files" / "data" / "bouwblokken.gobevents"
    importer = EventsProcessor([gebieden_schema], tconn, local_metadata=local_metadata)
    importer.load_events_from_file(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM gebieden_bouwblokken")]
    assert len(records) == 2

    # Import relation table row event for relation with shortname
    events_path = here / "files" / "data" / "bouwblokken_ligt_in_buurt_met_te_lange_naam.gobevents"
    importer.load_events_from_file(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM gebieden_bouwblokken")]
    assert len(records) == 2

    record = [r for r in records if r["identificatie"] == "03630012096976"][0]
    assert record["lgt_in_brt_id"] == "03630023754008ADD.1"
    assert record["lgt_in_brt_identificatie"] == "03630023754008ADD"
    assert record["lgt_in_brt_volgnummer"] == 1


def test_events_process_relation_without_table_update_parent_table(
    here, db_schema, tconn, local_metadata, brk_schema_without_bag_relations
):
    # First, verify that the relation table indeed does not exist
    res = next(
        tconn.execute(
            "SELECT EXISTS (SELECT FROM pg_tables WHERE schemaname = 'public' "
            "AND tablename  = 'brk_tenaamstellingen_van_kadastraalsubject');"
        )
    )
    assert res[0] is False

    importer = EventsProcessor(
        [brk_schema_without_bag_relations], tconn, local_metadata=local_metadata
    )

    # Import relation vanKadastraalsubject. Has no table, but we want to update the parent table
    events_path = here / "files" / "data" / "tenaamstellingen.gobevents"
    importer.load_events_from_file(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM brk_tenaamstellingen")]
    assert len(records) == 2
    assert [(r["identificatie"], r["van_kadastraalsubject_id"]) for r in records] == [
        ("NL.IMKAD.Tenaamstelling.ajdkfl4j4", "NL.IMKAD.Persoon.1124ji44kd"),
        ("NL.IMKAD.Tenaamstelling.adkfkadfkjld", "NL.IMKAD.Persoon.2f4802kkdd"),
    ]

    # Test that parent table is updated
    events_path = here / "files" / "data" / "tenaamstellingen_van_kadastraalsubject.gobevents"
    importer.load_events_from_file(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM brk_tenaamstellingen")]
    assert len(records) == 2
    assert [(r["identificatie"], r["van_kadastraalsubject_id"]) for r in records] == [
        ("NL.IMKAD.Tenaamstelling.ajdkfl4j4", "NL.IMKAD.Persoon.20042004eeeeefjd"),
        ("NL.IMKAD.Tenaamstelling.adkfkadfkjld", None),
    ]


def test_events_process_relation_without_table_update_parent_table_full_load(
    here, db_schema, tconn, local_metadata, brk_schema_without_bag_relations
):
    # First, verify that the relation table indeed does not exist
    res = next(
        tconn.execute(
            "SELECT EXISTS (SELECT FROM pg_tables WHERE schemaname = 'public' "
            "AND tablename  = 'brk_tenaamstellingen_van_kadastraalsubject');"
        )
    )
    assert res[0] is False

    importer = EventsProcessor(
        [brk_schema_without_bag_relations], tconn, local_metadata=local_metadata
    )

    # Import relation vanKadastraalsubject. Has no table, but we want to update the parent table
    events_path = here / "files" / "data" / "tenaamstellingen.gobevents"
    _load_events_from_file_as_full_load_sequence(importer, events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM brk_tenaamstellingen")]
    assert len(records) == 2
    assert [(r["identificatie"], r["van_kadastraalsubject_id"]) for r in records] == [
        ("NL.IMKAD.Tenaamstelling.ajdkfl4j4", "NL.IMKAD.Persoon.1124ji44kd"),
        ("NL.IMKAD.Tenaamstelling.adkfkadfkjld", "NL.IMKAD.Persoon.2f4802kkdd"),
    ]

    # Test that parent table is updated
    events = [
        (
            {
                "event_type": "ADD",
                "event_id": 1,
                "dataset_id": "brk",
                "table_id": "tenaamstellingen_vanKadastraalsubject",
                "full_load_sequence": True,
                "first_of_sequence": True,
                "last_of_sequence": False,
            },
            {
                "id": 1,
                "tenaamstellingen_id": "NL.IMKAD.Tenaamstelling.ajdkfl4j4.4",
                "tenaamstellingen_identificatie": "NL.IMKAD.Tenaamstelling.ajdkfl4j4",
                "tenaamstellingen_volgnummer": 4,
                "van_kadastraalsubject_id": "NL.IMKAD.Persoon.20042004eeeeefjd",
                "van_kadastraalsubject_identificatie": "NL.IMKAD.Persoon.20042004eeeeefjd",
            },
        ),
        (
            {
                "event_type": "ADD",
                "event_id": 2,
                "dataset_id": "brk",
                "table_id": "tenaamstellingen_vanKadastraalsubject",
                "full_load_sequence": True,
                "first_of_sequence": False,
                "last_of_sequence": True,
            },
            {
                "id": 3,
                "tenaamstellingen_id": "NL.IMKAD.Tenaamstelling.ajdkeeeefad.4",
                "tenaamstellingen_identificatie": "NL.IMKAD.Tenaamstelling.ajdkeeeefad",
                "tenaamstellingen_volgnummer": 4,
                "van_kadastraalsubject_id": "NL.IMKAD.Persoon.20042004eeeeefjd",
                "van_kadastraalsubject_identificatie": "NL.IMKAD.Persoon.20042004eeeeefjd",
            },
        ),
    ]

    importer.process_events(events, recovery_mode=False)
    records = [dict(r) for r in tconn.execute("SELECT * FROM brk_tenaamstellingen")]
    assert len(records) == 2

    # Full load does not update parent table, so result should be the same as above. This
    # testcase is included though to make sure we don't get an error on the missing relation table.
    assert [(r["identificatie"], r["van_kadastraalsubject_id"]) for r in records] == [
        ("NL.IMKAD.Tenaamstelling.ajdkfl4j4", "NL.IMKAD.Persoon.1124ji44kd"),
        ("NL.IMKAD.Tenaamstelling.adkfkadfkjld", "NL.IMKAD.Persoon.2f4802kkdd"),
    ]


def test_event_process_full_load_sequence(
    here, db_schema, tconn, local_metadata, nap_schema, gebieden_schema
):
    """Test consists of three parts:

    - First, load usual peilmerken events from peilmerken.gobevents.
    - Then, load events from peilmerken_full_load_sequence_start.gobevents. This should
      create the tmp table and not change the active table.
    - Lastly, load events from peilmerken_full_load_sequence_end.gobevents. This should
      end the full load sequence and
      replace the active table with the tmp table. Also, the tmp table should be removed.

    The peilmerken identificaties in the *.gobevents files used are unique, so we can
    check that the objects are in the expected tables.
    """

    def load_events(events_file):
        events_path = here / "files" / "data" / events_file
        importer = EventsProcessor(
            [nap_schema, gebieden_schema], tconn, local_metadata=local_metadata
        )
        importer.load_events_from_file_using_bulk(events_path)

    # 1.
    load_events("peilmerken.gobevents")
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]
    assert len(records) == 1
    assert records[0]["identificatie"] == "70780001"

    # 2.
    load_events("peilmerken_full_load_sequence_start.gobevents")
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]
    assert len(records) == 1
    assert records[0]["identificatie"] == "70780001"

    records = [dict(r) for r in tconn.execute("SELECT * FROM nap.nap_peilmerken_full_load")]
    assert len(records) == 3
    assert records[0]["identificatie"] == "70780002"
    assert records[1]["identificatie"] == "70780003"
    assert records[2]["identificatie"] == "70780004"

    # 3.
    load_events("peilmerken_full_load_sequence_end.gobevents")
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]
    assert len(records) == 4
    assert records[0]["identificatie"] == "70780002"
    assert records[1]["identificatie"] == "70780003"
    assert records[2]["identificatie"] == "70780004"
    assert records[3]["identificatie"] == "70780005"

    res = next(
        tconn.execute(
            "SELECT EXISTS (SELECT FROM pg_tables WHERE schemaname = 'public' "
            "AND tablename  = 'nap_peilmerken_full_load');"
        )
    )
    assert res[0] is False


def test_event_process_geometry_attrs(
    here, db_schema, tconn, local_metadata, brk_schema_without_bag_relations
):
    events_path = here / "files" / "data" / "kadastraleobjecten_geometry.gobevents"

    importer = EventsProcessor(
        [brk_schema_without_bag_relations], tconn, local_metadata=local_metadata
    )
    importer.load_events_from_file_using_bulk(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM brk_kadastraleobjecten")]

    assert len(records) == 1
    record = records[0]
    assert record["bijpijling_geometrie"] is not None
    assert record["geometrie"] is not None


def _create_peilmerken_event(id: str, jaar: int, type: str = "ADD", **kwargs) -> tuple[dict, dict]:
    """Creates a peilmerken event. Use 'jaar' to make sure the event is unique."""

    return (
        {
            "event_type": type,
            "dataset_id": "nap",
            "table_id": "peilmerken",
            **kwargs,
        },
        {
            "identificatie": id,
            "hoogte_tov_nap": -2.6954,
            "jaar": jaar,
            "merk": {
                "code": "7",
                "omschrijving": "Bijzondere merktekens bijvoorbeeld zeskantige bout, "
                "stalen pen etc.",
            },
            "omschrijving": "Gemaal aan de Ringvaart gelegen aan de Schipholweg. Vervallen?",
            "windrichting": "Z",
            "x_coordinaat_muurvlak": 84,
            "y_coordinaat_muurvlak": 37,
            "rws_nummer": "25D0039",
            "geometrie": "01010000204071000000000000C05FFC400000000050601D41",
            "status": {"code": 3, "omschrijving": "Vervallen"},
            "vervaldatum": "2018-04-23",
            "ligt_in_bouwblok_id": None,
            "ligt_in_bouwblok_identificatie": None,
            "ligt_in_bouwblok_volgnummer": None,
            "publiceerbaar": False,
        },
    )


def _assert_have_peilmerken(
    id_years: list[tuple[str, int]], tconn, check_full_load_table: bool = False
):
    tablename = "nap_peilmerken" if not check_full_load_table else "nap.nap_peilmerken_full_load"
    records = [dict(r) for r in tconn.execute(f"SELECT * FROM {tablename}")]  # noqa: S608
    assert len(records) == len(id_years)

    for record, (id, year) in zip(records, id_years):
        assert record["identificatie"] == id
        assert record["jaar"] == year


def _import_assert_result_expect_exception(
    importer, events, expected_result, recovery_mode=False, check_full_load_table=False
):
    with pytest.raises(Exception):
        importer.process_events(events, recovery_mode)
    _assert_have_peilmerken(expected_result, importer.conn, check_full_load_table)


def _import_assert_result(
    importer, events, expected_result, recovery_mode=False, check_full_load_table=False
):
    importer.process_events(events, recovery_mode)
    _assert_have_peilmerken(expected_result, importer.conn, check_full_load_table)


def test_event_process_recovery_regular(
    db_schema, engine, local_metadata, nap_schema, gebieden_schema
):
    """Tests adding of regular events (that are not part of a full load sequence) with
    and without recovery mode enabled.

    Initialise database with some rows.

    Try to add rows:
    - 1. One that already exists, with recovery_mode = False. Should raise an error.
    - 2. One that doesn't exist, with recovery_mode = False. Should be added.
    - 3. One that already exists, with recovery_mode = True. Should be ignored.
    - 4. One that doesn't exist, with recovery_mode = True. Should be added.
    """
    with engine.connect() as conn:
        importer = EventsProcessor(
            [nap_schema, gebieden_schema], conn, local_metadata=local_metadata
        )

        # Init
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event("1", 2018, event_id=1),
                _create_peilmerken_event("2", 2019, event_id=2),
            ],
            [("1", 2018), ("2", 2019)],
        )

        # 1. One that already exists, with recovery_mode = False. Should raise an error.
        _import_assert_result_expect_exception(
            importer, [_create_peilmerken_event("1", 2017, event_id=3)], [("1", 2018), ("2", 2019)]
        )

        # 2. One that doesn't exist, with recovery_mode = False. Should be added.
        _import_assert_result(
            importer,
            [_create_peilmerken_event("3", 2020, event_id=4)],
            [("1", 2018), ("2", 2019), ("3", 2020)],
        )

        # 3. One that already exists, with recovery_mode = True. Should be ignored.
        _import_assert_result(
            importer,
            [_create_peilmerken_event("3", 2019, event_id=5)],
            [("1", 2018), ("2", 2019), ("3", 2020)],
            recovery_mode=True,
        )

        # 4. One that doesn't exist, with recovery_mode = True. Should be added.
        _import_assert_result(
            importer,
            [_create_peilmerken_event("4", 2021, event_id=6)],
            [("1", 2018), ("2", 2019), ("3", 2020), ("4", 2021)],
            recovery_mode=True,
        )
        conn.close()


def test_event_process_recovery_full_load_first(
    db_schema, engine, local_metadata, nap_schema, gebieden_schema
):
    """Tests adding of events that are part of a full load sequence and where the first
    event is the first in the sequence.

    Initialise full load with some rows.

    Try to add multiple rows, where the first one:
    - Already exists, with recovery_mode = False. New rows should be added.
    - Doesn't exist, with recovery_mode = False. New rows should be added.
    - Already exists, with recovery_mode = True. New rows should be added.
    - Doesn't exist, with recovery_mode = True. New rows should be added.

    Note: When the first event to handle in this case is the first in the sequence, the
    _full_load table is always truncated. This means that in every case we replace the
    table.

    """
    with engine.connect() as conn:
        importer = EventsProcessor(
            [nap_schema, gebieden_schema], conn, local_metadata=local_metadata
        )

        # Init
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event(
                    "1", 2018, full_load_sequence=True, first_of_sequence=True, event_id=1
                ),
                _create_peilmerken_event("2", 2019, full_load_sequence=True, event_id=2),
            ],
            [("1", 2018), ("2", 2019)],
            check_full_load_table=True,
        )

        # 1. Already exists, with recovery_mode = False. New rows should be added.
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event(
                    "1", 2020, full_load_sequence=True, first_of_sequence=True, event_id=3
                ),
                _create_peilmerken_event("2", 2021, full_load_sequence=True, event_id=4),
            ],
            [("1", 2020), ("2", 2021)],
            check_full_load_table=True,
        )

        # 2. Doesn't exist, with recovery_mode = False. New rows should be added.
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event(
                    "3", 2020, full_load_sequence=True, first_of_sequence=True, event_id=5
                ),
                _create_peilmerken_event("4", 2021, full_load_sequence=True, event_id=6),
            ],
            [("3", 2020), ("4", 2021)],
            check_full_load_table=True,
        )

        # 3. Already exists, with recovery_mode = True. New rows should be added.
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event(
                    "3", 2022, full_load_sequence=True, first_of_sequence=True, event_id=7
                ),
                _create_peilmerken_event("4", 2023, full_load_sequence=True, event_id=8),
            ],
            [("3", 2022), ("4", 2023)],
            check_full_load_table=True,
            recovery_mode=True,
        )

        # 4. Doesn't exist, with recovery_mode = True. New rows should be added.
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event(
                    "5", 2020, full_load_sequence=True, first_of_sequence=True, event_id=9
                ),
                _create_peilmerken_event("6", 2021, full_load_sequence=True, event_id=10),
            ],
            [("5", 2020), ("6", 2021)],
            check_full_load_table=True,
            recovery_mode=True,
        )


def test_event_process_recovery_full_load_no_first_no_last(
    db_schema, engine, local_metadata, nap_schema, gebieden_schema
):
    """Tests adding of events that are part of a full load sequence and where events to
    be added are neither the first nor the last in the sequence.

    Initialise full load with some rows.

    Try to add multiple rows, where the first one:
    - Already exists, with recovery_mode = False. Should raise an error.
    - Doesn't exist, with recovery_mode = False. New rows should be added.
    - Already exists, with recovery_mode = True. Rows should be ignored.
    - Doesn't exist, with recovery_mode = True. New rows should be added.

    """
    with engine.connect() as conn:
        importer = EventsProcessor(
            [nap_schema, gebieden_schema], conn, local_metadata=local_metadata
        )

        # Init
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event(
                    "1", 2018, full_load_sequence=True, first_of_sequence=True, event_id=1
                ),
                _create_peilmerken_event("2", 2019, full_load_sequence=True, event_id=2),
            ],
            [("1", 2018), ("2", 2019)],
            check_full_load_table=True,
        )

        # 1. Already exists, with recovery_mode = False. Should raise an error.
        _import_assert_result_expect_exception(
            importer,
            [
                _create_peilmerken_event("1", 2022, full_load_sequence=True, event_id=3),
                _create_peilmerken_event("2", 2023, full_load_sequence=True, event_id=4),
            ],
            [("1", 2018), ("2", 2019)],
            check_full_load_table=True,
        )

        # 2. Doesn't exist, with recovery_mode = False. New rows should be added.
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event("3", 2020, full_load_sequence=True, event_id=5),
                _create_peilmerken_event("4", 2021, full_load_sequence=True, event_id=6),
            ],
            [("1", 2018), ("2", 2019), ("3", 2020), ("4", 2021)],
            check_full_load_table=True,
        )

        # 3. Already exists, with recovery_mode = True. Rows should be ignored.
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event("1", 2022, full_load_sequence=True, event_id=7),
                _create_peilmerken_event("2", 2023, full_load_sequence=True, event_id=8),
            ],
            [("1", 2018), ("2", 2019), ("3", 2020), ("4", 2021)],
            check_full_load_table=True,
            recovery_mode=True,
        )

        # 4. Doesn't exist, with recovery_mode = True. New rows should be added.
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event("5", 2022, full_load_sequence=True, event_id=9),
                _create_peilmerken_event("6", 2023, full_load_sequence=True, event_id=10),
            ],
            [("1", 2018), ("2", 2019), ("3", 2020), ("4", 2021), ("5", 2022), ("6", 2023)],
            check_full_load_table=True,
            recovery_mode=True,
        )


def test_event_process_recovery_full_load_last_table_empty(
    db_schema, engine, local_metadata, nap_schema, gebieden_schema
):
    """Tests adding of events that are part of a full load sequence and where the last
    event is the last in the sequence and the _full_load table is empty.

    No full load initialisation; table should be empty

    Try to add multiple rows, where the first one:
    - Doesn't exist, with recovery_mode = False. New rows should be added and object table
      replaced.
    - Doesn't exist, with recovery_mode = True. Everything in this message should be ignored

    """
    with engine.connect() as conn:
        importer = EventsProcessor(
            [nap_schema, gebieden_schema], conn, local_metadata=local_metadata
        )

        def _init_empty_full_load_table(importer):
            # Init and truncate right after to get inconsistent state
            _import_assert_result(
                importer,
                [
                    _create_peilmerken_event(
                        "1", 2018, full_load_sequence=True, first_of_sequence=True, event_id=1
                    ),
                    _create_peilmerken_event("2", 2019, full_load_sequence=True, event_id=2),
                ],
                [("1", 2018), ("2", 2019)],
                check_full_load_table=True,
            )
            importer.conn.execute("TRUNCATE TABLE nap.nap_peilmerken_full_load")

        # 1. Doesn't exist, with recovery_mode = False. New rows should be added and object
        #    table replaced.
        _init_empty_full_load_table(importer)
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event("1", 2018, full_load_sequence=True, event_id=3),
                _create_peilmerken_event(
                    "2", 2019, full_load_sequence=True, last_of_sequence=True, event_id=4
                ),
            ],
            [
                ("1", 2018),
                ("2", 2019),
            ],
        )

        # 2. Doesn't exist, with recovery_mode = True. Everything in this message should
        #    be ignored
        _init_empty_full_load_table(importer)
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event("3", 2018, full_load_sequence=True, event_id=5),
                _create_peilmerken_event(
                    "4", 2019, full_load_sequence=True, last_of_sequence=True, event_id=6
                ),
            ],
            [
                ("1", 2018),
                ("2", 2019),
            ],
            recovery_mode=True,
        )


def test_event_process_recovery_full_load_last_table_not_empty(
    db_schema, engine, local_metadata, nap_schema, gebieden_schema
):
    """Tests adding of events that are part of a full load sequence and where the last event is
    the last in the sequence and the _full_load table is not empty.

    Initialise full load with some rows.

    Try to add multiple rows, where the first one:
    - Already exists, with recovery_mode = False. Should raise an error.
    - Doesn't exist, with recovery_mode = False. New rows should be added and object table
      replaced.
    - Already exists, with recovery_mode = True. Rows should be ignored. Object table replaced.
    - Doesn't exist, with recovery_mode = True. Rows should be added. Object table replaced.

    """
    with engine.connect() as conn:
        importer = EventsProcessor(
            [nap_schema, gebieden_schema], conn, local_metadata=local_metadata
        )

        def _init_full_load_table(importer):
            for table in ["nap_peilmerken_full_load", "benk_lasteventids"]:
                try:
                    conn.execute(f"TRUNCATE TABLE {table}")  # noqa: S608
                except Exception:  # noqa: S110
                    pass
            importer.lasteventids.clear_cache()

            _import_assert_result(
                importer,
                [
                    _create_peilmerken_event(
                        "1", 2018, full_load_sequence=True, first_of_sequence=True, event_id=1
                    ),
                    _create_peilmerken_event("2", 2019, full_load_sequence=True, event_id=2),
                ],
                [("1", 2018), ("2", 2019)],
                check_full_load_table=True,
            )

        # 1. Already exists, with recovery_mode = False. Should raise an error.
        _init_full_load_table(importer)
        _import_assert_result_expect_exception(
            importer,
            [
                _create_peilmerken_event("1", 2022, full_load_sequence=True, event_id=3),
                _create_peilmerken_event(
                    "2", 2023, full_load_sequence=True, last_of_sequence=True, event_id=4
                ),
            ],
            [("1", 2018), ("2", 2019)],
            check_full_load_table=True,
        )

        # 2. Doesn't exist, with recovery_mode = False. New rows should be added and
        #    object table replaced.
        _init_full_load_table(importer)
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event("3", 2020, full_load_sequence=True, event_id=5),
                _create_peilmerken_event(
                    "4", 2021, full_load_sequence=True, last_of_sequence=True, event_id=6
                ),
            ],
            [("1", 2018), ("2", 2019), ("3", 2020), ("4", 2021)],
        )

        # 3. Already exists, with recovery_mode = True. Rows should be ignored. Object
        #    table replaced.
        _init_full_load_table(importer)
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event("1", 2023, full_load_sequence=True, event_id=7),
                _create_peilmerken_event(
                    "2", 2024, full_load_sequence=True, last_of_sequence=True, event_id=8
                ),
            ],
            [("1", 2018), ("2", 2019)],
            recovery_mode=True,
        )

        # 4. Doesn't exist, with recovery_mode = True. Rows should be added. Object table
        #    replaced.
        _init_full_load_table(importer)
        _import_assert_result(
            importer,
            [
                _create_peilmerken_event("3", 2020, full_load_sequence=True, event_id=9),
                _create_peilmerken_event(
                    "4", 2021, full_load_sequence=True, last_of_sequence=True, event_id=10
                ),
            ],
            [("1", 2018), ("2", 2019), ("3", 2020), ("4", 2021)],
            recovery_mode=True,
        )


def test_event_process_last_event_id(
    here, db_schema, tconn, local_metadata, nap_schema, gebieden_schema, benk_schema
):
    def get_last_event_id(tablename: str = "nap_peilmerken"):
        res = tconn.execute(
            f"SELECT last_event_id FROM benk_lasteventids "  # noqa: S608  # nosec: B608
            f"WHERE \"table\"='{tablename}'"
        ).fetchone()

        return res[0] if res is not None else None

    importer = EventsProcessor(
        [nap_schema, gebieden_schema, benk_schema], tconn, local_metadata=local_metadata
    )

    # 1. Assert start state
    assert get_last_event_id() is None

    events = [
        _create_peilmerken_event("1", 2018, event_id=203),
        _create_peilmerken_event("2", 2019, event_id=210),
    ]
    importer.process_events(events)

    # 2. Add rows and assert they exist
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]
    assert [2018, 2019] == [r["jaar"] for r in records]
    assert get_last_event_id() == 210

    events = [
        _create_peilmerken_event("2", 2020, type="MODIFY", event_id=211),
    ]
    importer.process_events(events)

    # 3. Assert event with newer ID is applied
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]
    assert [2018, 2020] == [r["jaar"] for r in records]
    assert get_last_event_id() == 211

    events = [
        _create_peilmerken_event("1", 2021, type="MODIFY", event_id=204),
        _create_peilmerken_event("2", 2021, type="MODIFY", event_id=211),
    ]
    importer.process_events(events)

    # 4. Assert event with older IDs are ignored
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]
    assert [2018, 2020] == [r["jaar"] for r in records]
    assert get_last_event_id() == 211


def test_event_process_last_event_id_full_load_sequence(
    here, db_schema, tconn, local_metadata, nap_schema, gebieden_schema, benk_schema
):
    def get_last_event_id(tablename: str = "nap_peilmerken"):
        res = tconn.execute(
            f"SELECT last_event_id FROM benk_lasteventids "  # noqa: S608
            f"WHERE \"table\"='{tablename}'"  # noqa: S608
        ).fetchone()

        return res[0] if res is not None else None

    importer = EventsProcessor(
        [nap_schema, gebieden_schema, benk_schema], tconn, local_metadata=local_metadata
    )

    # 1. Assert start state
    assert get_last_event_id("nap_peilmerken") is None
    assert get_last_event_id("nap_peilmerken_full_load") is None

    events = [
        _create_peilmerken_event(
            "1", 2018, event_id=203, full_load_sequence=True, first_of_sequence=True
        ),
        _create_peilmerken_event("2", 2019, event_id=210, full_load_seuqence=True),
    ]
    importer.process_events(events)

    # 2. Add rows and assert they exist
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap.nap_peilmerken_full_load")]
    assert [2018, 2019] == [r["jaar"] for r in records]
    assert get_last_event_id("nap_peilmerken") is None
    assert get_last_event_id("nap_peilmerken_full_load") == 210

    events = [
        _create_peilmerken_event("3", 2020, type="ADD", event_id=212, full_load_sequence=True),
    ]
    importer.process_events(events)

    # 3. Assert event with newer ID is applied
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap.nap_peilmerken_full_load")]
    assert [2018, 2019, 2020] == [r["jaar"] for r in records]
    assert get_last_event_id("nap_peilmerken") is None
    assert get_last_event_id("nap_peilmerken_full_load") == 212

    events = [
        _create_peilmerken_event("4", 2021, type="ADD", event_id=204, full_load_sequence=True),
        _create_peilmerken_event("5", 2021, type="ADD", event_id=211, full_load_sequence=True),
    ]
    importer.process_events(events)

    # 4. Assert event with older IDs are ignored
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap.nap_peilmerken_full_load")]
    assert [2018, 2019, 2020] == [r["jaar"] for r in records]
    assert get_last_event_id("nap_peilmerken") is None
    assert get_last_event_id("nap_peilmerken_full_load") == 212

    # 5. End full load. Table should be replaced and last_event_id copied to main table and reset
    events = [
        _create_peilmerken_event("4", 2021, type="ADD", event_id=213, full_load_sequence=True),
        _create_peilmerken_event(
            "5", 2022, type="ADD", event_id=217, full_load_sequence=True, last_of_sequence=True
        ),
    ]
    importer.process_events(events)

    # 4. Assert event with older IDs are ignored
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]
    assert [2018, 2019, 2020, 2021, 2022] == [r["jaar"] for r in records]
    assert get_last_event_id("nap_peilmerken") == 217
    assert get_last_event_id("nap_peilmerken_full_load") is None


def test_events_process_full_load_sequence_snake_cased_schema(
    here, db_schema, tconn, local_metadata, brk2_simple_schema
):
    """Tests whether the correct (snake_cased) schema for brk2 is used for the full load."""

    event_meta = {
        "event_type": "ADD",
        "event_id": 1,
        "dataset_id": "brk2",
        "table_id": "gemeentes",
        "full_load_sequence": True,
        "first_of_sequence": True,
    }
    event_data = {"identificatie": "0363", "naam": "Amsterdam"}

    importer = EventsProcessor([brk2_simple_schema], tconn, local_metadata=local_metadata)
    importer.process_event(event_meta, event_data)

    records = [dict(r) for r in tconn.execute("SELECT * FROM brk_2.brk_2_gemeentes_full_load")]
    assert len(records) == 1
    assert records[0]["identificatie"] == "0363"
    assert records[0]["naam"] == "Amsterdam"

    event_meta = {
        "event_type": "ADD",
        "event_id": 2,
        "dataset_id": "brk2",
        "table_id": "gemeentes",
        "full_load_sequence": True,
        "last_of_sequence": True,
    }
    event_data = {
        "identificatie": "0457",
        "naam": "Weesp",
    }
    importer.process_event(event_meta, event_data)

    records = [dict(r) for r in tconn.execute("SELECT * FROM brk_2_gemeentes")]
    assert len(records) == 2
    assert records[0]["identificatie"] == "0363"
    assert records[0]["naam"] == "Amsterdam"
    assert records[1]["identificatie"] == "0457"
    assert records[1]["naam"] == "Weesp"


def test_events_process_full_load_relation_update_parent_table(
    here, db_schema, tconn, local_metadata, nap_schema, gebieden_schema
):
    events_path = here / "files" / "data" / "peilmerken.gobevents"
    importer = EventsProcessor([nap_schema, gebieden_schema], tconn, local_metadata=local_metadata)
    importer.load_events_from_file_using_bulk(events_path)
    records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]

    # Imported objects without relations
    assert len(records) == 1
    assert records[0]["ligt_in_bouwblok_id"] is None
    assert records[0]["ligt_in_bouwblok_identificatie"] is None
    assert records[0]["ligt_in_bouwblok_volgnummer"] is None

    event_meta = {
        "event_type": "ADD",
        "event_id": 1,
        "dataset_id": "nap",
        "table_id": "peilmerken_ligtInBouwblok",
        "full_load_sequence": True,
        "first_of_sequence": True,
        "last_of_sequence": True,
    }
    event_data = {
        "id": 1,
        "peilmerken_id": "70780001",
        "peilmerken_identificatie": "70780001",
        "ligt_in_bouwblok_id": "03630012095746.1",
        "ligt_in_bouwblok_identificatie": "03630012095746",
        "ligt_in_bouwblok_volgnummer": 1,
    }

    importer.process_event(event_meta, event_data)

    rel_records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken_ligt_in_bouwblok")]
    parent_records = [dict(r) for r in tconn.execute("SELECT * FROM nap_peilmerken")]

    # Should have updated relation columns in  parent (object) table
    assert len(rel_records) == 1
    assert len(parent_records) == 1

    assert parent_records[0]["ligt_in_bouwblok_id"] == "03630012095746.1"
    assert parent_records[0]["ligt_in_bouwblok_identificatie"] == "03630012095746"
    assert parent_records[0]["ligt_in_bouwblok_volgnummer"] == 1


def load_json_results_file(location):
    class _JSONDecoder(json.JSONDecoder):
        """Custom JSON decoder that converts date(time) strings to date(time) objects."""

        def __init__(self, *args, **kwargs):
            json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

        def object_hook(self, obj):
            ret = {}
            for key, value in obj.items():
                if key in ("begin_geldigheid", "eind_geldigheid") and value is not None:
                    if len(value) == 10:
                        ret[key] = date.fromisoformat(value)
                    else:
                        ret[key] = datetime.fromisoformat(value)
                else:
                    ret[key] = value
            return ret

    with open(location) as f:
        return json.load(f, cls=_JSONDecoder)


def delete_id(d: dict):
    return {k: v for k, v in d.items() if k != "id"}


def delete_ids(lst: list[dict]):
    return [delete_id(d) for d in lst]


def assert_results(tconn, expected_results: dict, testname: str):
    for table_name, expected_result in expected_results.items():
        records = [dict(r) for r in tconn.execute(f"SELECT * FROM {table_name}")]  # noqa: S608
        records = delete_ids(records)

        assert len(records) == len(
            expected_result
        ), f"Number of records in {table_name} does not match for test {testname}"
        for res in expected_result:
            assert res in records, f"Record {res} not found in {table_name} for test {testname}"
        for rec in records:
            assert (
                rec in expected_result
            ), f"Unexpected record {rec} found in {table_name} for test {testname}"


def test_events_nested_table(here, db_schema, tconn, local_metadata, bag_verblijfsobjecten_schema):
    expected_results = load_json_results_file(
        here / "files" / "data" / "expect" / "events_nested_table.json"
    )

    importer = EventsProcessor(
        [bag_verblijfsobjecten_schema], tconn, local_metadata=local_metadata
    )

    # Load initial data with nested objects
    events_path = here / "files" / "data" / "verblijfsobjecten.gobevents"
    importer.load_events_from_file(events_path)
    assert_results(tconn, expected_results["initial_add"], "Load initial data")

    # Modify nested objects
    events_path = here / "files" / "data" / "verblijfsobjecten.modify_nested.gobevents"
    importer.load_events_from_file(events_path)
    assert_results(tconn, expected_results["modify_nested"], "Modify nested objects")

    # Empty nested objects
    events_path = here / "files" / "data" / "verblijfsobjecten.empty_nested.gobevents"
    importer.load_events_from_file(events_path)
    assert_results(tconn, expected_results["empty_nested"], "Remove nested objects")

    # Modify nested objects again
    events_path = here / "files" / "data" / "verblijfsobjecten.modify_nested_2.gobevents"
    importer.load_events_from_file(events_path)
    assert_results(tconn, expected_results["modify_nested"], "Modify nested objects again")

    # Delete full object
    events_path = here / "files" / "data" / "verblijfsobjecten.delete.gobevents"
    importer.load_events_from_file(events_path)
    assert_results(tconn, expected_results["delete"], "Delete everything")

    # Now test full load. Full load only works with ADD events, so we reuse the ADD event from
    # above and add the full load metadata.
    events_path = here / "files" / "data" / "verblijfsobjecten.gobevents"
    _load_events_from_file_as_full_load_sequence(importer, events_path)

    assert_results(tconn, expected_results["initial_add"], "Load initial data using full load")


def test_full_load_shortnames(here, db_schema, tconn, local_metadata, hr_simple_schema):
    importer = EventsProcessor([hr_simple_schema], tconn, local_metadata=local_metadata)

    # First import an object with nested objects
    events = [
        (
            {
                "event_type": "ADD",
                "event_id": 1,
                "dataset_id": "hr",
                "table_id": "maatschappelijkeactiviteiten",
                "full_load_sequence": True,
                "first_of_sequence": True,
                "last_of_sequence": True,
            },
            {
                "kvknummer": 42,
                "email_adressen": [
                    {
                        "email_adres": "address1@example.com",
                    },
                    {
                        "email_adres": "address2@example.com",
                    },
                ],
            },
        )
    ]
    importer.process_events(events)

    # Not testing contents here, but merely the fact that the right tables are used without errors
    records = [dict(r) for r in tconn.execute("SELECT * FROM hr_mac")]
    assert len(records) == 1
    assert records[0]["heeft_hoofdvestiging_id"] is None

    nested_records = [dict(r) for r in tconn.execute("SELECT * FROM hr_mac_email_adressen")]
    assert len(nested_records) == 2
    assert nested_records[0]["parent_id"] == "42"
    assert nested_records[0]["email_adres"] == "address1@example.com"

    # Now test adding a relation object that references a parent table with short name
    events = [
        (
            {
                "dataset_id": "hr",
                "table_id": "maatschappelijkeactiviteiten_heeftHoofdvestiging",
                "event_type": "ADD",
                "event_id": 1658565091,
                "tid": "42.AMSBI.24902480",
                "generated_timestamp": "2023-10-05T09:59:05.314873",
                "full_load_sequence": True,
                "first_of_sequence": True,
                "last_of_sequence": True,
            },
            {
                "mac_kvknummer": "42",
                "mac_id": "42",
                "heeft_hoofdvestiging_vestigingsnummer": "24902480",
                "heeft_hoofdvestiging_id": "24902480",
                "begin_geldigheid": None,
                "eind_geldigheid": None,
                "id": 457172,
            },
        )
    ]

    importer.process_events(events)
    rel_records = [dict(r) for r in tconn.execute("SELECT * FROM hr_mac_heeft_hoofdvestiging")]
    assert len(rel_records) == 1
    assert rel_records[0]["id"] == 457172
    assert rel_records[0]["mac_id"] == "42"
    assert rel_records[0]["heeft_hoofdvestiging_id"] == "24902480"

    records = [dict(r) for r in tconn.execute("SELECT * FROM hr_mac")]
    assert len(records) == 1
    assert records[0]["heeft_hoofdvestiging_id"] == "24902480"


def test_full_load_shortnames_update(here, db_schema, tconn, local_metadata, hr_simple_schema):
    importer = EventsProcessor([hr_simple_schema], tconn, local_metadata=local_metadata)

    # First import an object with nested objects
    events = [
        (
            {
                "event_type": "ADD",
                "event_id": 1,
                "dataset_id": "hr",
                "table_id": "maatschappelijkeactiviteiten",
            },
            {
                "kvknummer": 42,
                "email_adressen": [
                    {
                        "email_adres": "address1@example.com",
                    },
                    {
                        "email_adres": "address2@example.com",
                    },
                ],
            },
        )
    ]
    importer.process_events(events)

    # Not testing contents here, but merely the fact that the right tables are used without errors
    records = [dict(r) for r in tconn.execute("SELECT * FROM hr_mac")]
    assert len(records) == 1
    assert records[0]["heeft_hoofdvestiging_id"] is None

    nested_records = [dict(r) for r in tconn.execute("SELECT * FROM hr_mac_email_adressen")]
    assert len(nested_records) == 2
    assert nested_records[0]["parent_id"] == "42"
    assert nested_records[0]["email_adres"] == "address1@example.com"

    # Now test adding a relation object that references a parent table with short name
    events = [
        (
            {
                "dataset_id": "hr",
                "table_id": "maatschappelijkeactiviteiten_heeftHoofdvestiging",
                "event_type": "ADD",
                "event_id": 1658565091,
                "tid": "42.AMSBI.24902480",
                "generated_timestamp": "2023-10-05T09:59:05.314873",
            },
            {
                "mac_kvknummer": "42",
                "mac_id": "42",
                "heeft_hoofdvestiging_vestigingsnummer": "24902480",
                "heeft_hoofdvestiging_id": "24902480",
                "begin_geldigheid": None,
                "eind_geldigheid": None,
                "id": 457172,
            },
        )
    ]

    importer.process_events(events)
    rel_records = [dict(r) for r in tconn.execute("SELECT * FROM hr_mac_heeft_hoofdvestiging")]
    assert len(rel_records) == 1
    assert rel_records[0]["id"] == 457172
    assert rel_records[0]["mac_id"] == "42"
    assert rel_records[0]["heeft_hoofdvestiging_id"] == "24902480"

    records = [dict(r) for r in tconn.execute("SELECT * FROM hr_mac")]
    assert len(records) == 1
    assert records[0]["heeft_hoofdvestiging_id"] == "24902480"


def test_reset_lasteventid_after_incomplete_full_load(
    here, db_schema, tconn, local_metadata, nap_schema, gebieden_schema
):
    """This testcase tests whether the lasteventid is reset after an incomplete full load sequence.
    This should not happen during normal usage, but can happen when the full load stream is
    manually removed from the queue.
    """
    importer = EventsProcessor([nap_schema, gebieden_schema], tconn, local_metadata=local_metadata)
    events = [
        _create_peilmerken_event(
            "1", 2018, event_id=3, full_load_sequence=True, first_of_sequence=True
        ),
        _create_peilmerken_event(
            "2", 2018, event_id=4, full_load_sequence=True, first_of_sequence=False
        ),
    ]

    importer.process_events(events)
    lasteventrecord = tconn.execute(
        "SELECT * FROM benk_lasteventids WHERE \"table\" = 'nap_peilmerken_full_load'"
    ).fetchone()
    assert lasteventrecord["last_event_id"] == 4

    events = [
        _create_peilmerken_event(
            "1", 2018, event_id=1, full_load_sequence=True, first_of_sequence=True
        ),
    ]
    importer.process_events(events)
    lasteventrecord = tconn.execute(
        "SELECT * FROM benk_lasteventids WHERE \"table\" = 'nap_peilmerken_full_load'"
    ).fetchone()
    assert lasteventrecord["last_event_id"] == 1


def test_avoid_duplicate_key_after_full_load(
    here, db_schema, tconn, local_metadata, bag_verblijfsobjecten_schema
):
    """Make sure we don't get duplicate key errors after a full load sequence with a serial id
    field in the table."""

    def create_event(gebruiksdoel_cnt: int, event_id: int, identificatie: str, **extra_headers):
        gebruiksdoelen = [
            {"code": i, "omschrijving": f"doel {i}"} for i in range(1, gebruiksdoel_cnt + 1)
        ]
        return (
            {
                "event_type": "ADD",
                "event_id": event_id,
                "dataset_id": "bag",
                "table_id": "verblijfsobjecten",
                **extra_headers,
            },
            {
                "identificatie": identificatie,
                "volgnummer": 1,
                "gebruiksdoel": gebruiksdoelen,
                "toegang": None,
                "ligt_in_buurt": {},
                "begin_geldigheid": "2018-10-22T00:00:00.000000",
                "eind_geldigheid": None,
            },
        )

    importer = EventsProcessor(
        [bag_verblijfsobjecten_schema], tconn, local_metadata=local_metadata
    )

    # Add objects with in total 4 nested objects
    full_load_events = [
        create_event(2, 1, "VB1", full_load_sequence=True, first_of_sequence=True),
        create_event(2, 2, "VB2", full_load_sequence=True, last_of_sequence=True),
    ]
    importer.process_events(full_load_events)

    update_event = [create_event(1, 3, "VB3")]
    importer.process_events(update_event)



from __future__ import annotations

from factory import Faker
from faker.generator import Generator
from faker.providers import BaseProvider
from faker.providers.python import Provider as PyProvider


class NullableFloatProvider(BaseProvider):
    """Provider for a float or None provider.

    To be used for nullable fields in a database.

    """

    def __init__(self, generator: Generator) -> None:
        """Declare the other faker providers."""
        super().__init__(generator)
        self.python = PyProvider(generator)

    def nullable_float(  # noqa: D102
        self,
        left_digits=None,
        right_digits=None,
        positive=False,
        min_value=None,
        max_value=None,
        nullable=False,
    ) -> float | None:

        if nullable:
            if self.generator.random.randint(0, 1):
                return None

        return self.python.pyfloat(
            left_digits=left_digits,
            right_digits=right_digits,
            positive=positive,
            min_value=min_value,
            max_value=max_value,
        )


Faker.add_provider(NullableFloatProvider, locale="nl_NL")
